# -*- coding: utf-8 -*-
"""
Created on Thu Jan 19 14:56:45 2023

@author: jced0001
"""
import numpy as np

class Osci2T:
    """
    Nanonis Oscilloscope 2-Channels Module
    """
    def __init__(self,NanonisTCP):
        self.NanonisTCP = NanonisTCP

    def Run(self):
        """
        Starts the Oscilloscope 2-Channels Module
        
        This module does not run when its front panel is closed. To automate 
        measurements it is required to run the module first using this function
    

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.Run', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def ChsSet(self,chA=-1,chB=-1):
        """
        Sets the channels to display in the Oscilloscope 2-Channels. Leaving 
        chA or chB as -1 will leave the setting as is in Nanonis

        Parameters
        ----------
        chA : Sets the channel A, where chA is comprised between 0 and 23, and 
              it corresponds to the list of signals assigned to the 24 slots in
              the Signals Manager.
              To get the signal name and its corresponding index in the list of
              the 128 available signals in the nanonis controller use the 
              Signals.InSlotsGet function
        chB : sets the channel B where the in index is comprised between 0 and 
              23, and it corresponds to the list of signals assigned to the 24
              slots in the Signals Manager

        """
        chA_temp,chB_temp = self.ChsGet()
        if(chA == -1): chA = chA_temp
        if(chB == -1): chB = chB_temp
        
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.ChsSet', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(chA,4)
        hex_rep += self.NanonisTCP.to_hex(chB,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def ChsGet(self):
        """
        Returns the channels displayed in the Oscilloscope 2-Channels

        Returns
        -------
        chA : Returns channel A, where chA is comprised between 0 and 23, and 
              it corresponds to the list of signals assigned to the 24 slots in
              the Signals Manager.
              To get the signal name and its corresponding index in the list of
              the 128 available signals in the nanonis controller use the 
              Signals.InSlotsGet function
        chB : Returns channel B where the in index is comprised between 0 and 
              23, and it corresponds to the list of signals assigned to the 24
              slots in the Signals Manager

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.ChsGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        chA = self.NanonisTCP.hex_to_uint16(response[0:4])
        chB = self.NanonisTCP.hex_to_uint16(response[4:8])
        
        return [chA,chB]
    
    def TimebaseSet(self,timebaseIndex):
        """
        Sets the timebase in the Oscilloscope 2-Channels.
        To set the timebase, use the Osci2T.TimebaseGet function first to 
        obtain a list of available timebases. Then, use the index of the 
        desired timebase as input to this function.
        The available timebases depend on the RT frequency and the RT 
        oversampling.

        Parameters
        ----------
        timebaseIndex : Timebase index

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.TimebaseSet', body_size=2)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(timebaseIndex,2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def TimebaseGet(self):
        """
        Returns the timebase in the Oscilloscope 2-Channels.
        The available timebases depend on the RT frequency and the RT 
        oversampling.

        Returns
        -------
        timebaseIndex : Timebase index
        timebases : returns an array of the timebsaes values in seconds

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.TimebaseGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        timebaseIndex = self.NanonisTCP.hex_to_int16(response[0:2])
        numTimebases  = self.NanonisTCP.hex_to_int32(response[2:6])
        
        idx = 6
        timebases = []
        for n in range(numTimebases):
            timebases.append(self.NanonisTCP.hex_to_float32(response[idx:idx+4]))
            idx += 4
        
        return [timebaseIndex,np.array(timebases)]
    
    def OversamplSet(self,oversamplingIndex):
        """
        Sets the oversampling in the Oscilloscope 2-Channels

        Parameters
        ----------
        oversamplingIndex : defines how many integer number of samples each 
                            data point is averaged over. Index 0 means 50 
                            samples, index 1 means 20, index 2 means 10, index 
                            3 means 5, index 4 means 2, and index 5 means 1 
                            sample (so no averaging)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.OversamplSet', body_size=2)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(oversamplingIndex,2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def OversamplGet(self):
        """
        Returns the oversampling in the Oscilloscope 2-Channels

        Returns
        -------
        oversamplingIndex : defines how many integer number of samples each 
                            data point is averaged over. Index 0 means 50 
                            samples, index 1 means 20, index 2 means 10, index 
                            3 means 5, index 4 means 2, and index 5 means 1 
                            sample (so no averaging)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.OversamplGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(2)
        
        oversamplingIndex = self.NanonisTCP.hex_to_int16(response[0:2])
        
        return oversamplingIndex
    
    def TrigSet(self,mode,channel,slope,level,hysteresis,position):
        """
        Sets the trigger configuration in Oscilloscope 2-Channels

        Parameters
        ----------
        mode        : sets the trigger mode. For Immediate mode, no further 
                      config is required.
                      0 : Immediate mode
                      1 : Level mode
                      2 : Auto mode
              
        channel     : sets the channel used to trigger
                      0 : chA
                      1 : chB
        slope       : sets whether to trigger on rising or falling slope
                      0 : falling
                      1 : rising
        level       : sets the value the signal must cross (in the directin of
                      the slope specified) to trigger
        hysteresis  : is used to prevent noise from causing a false trigger.
                      For a rising edge trigger, the signal must pass below
                      (level - hysteresis) before a trigger level crossing is
                      detected
                      For a falling edge trigger, the signal must pass above
                      (level + hysteresis) before a trigger level crossing is
                      detected
        position    : sets the pre-sampling trigger position in seconds

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.TrigSet', body_size=30)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(mode,2)
        hex_rep += self.NanonisTCP.to_hex(channel,2)
        hex_rep += self.NanonisTCP.to_hex(slope,2)
        hex_rep += self.NanonisTCP.float64_to_hex(level)
        hex_rep += self.NanonisTCP.float64_to_hex(hysteresis)
        hex_rep += self.NanonisTCP.float64_to_hex(position)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def TrigGet(self):
        """
        

        Returns
        -------
        mode        : Returns the trigger mode.
                      0 : Immediate mode
                      1 : Level mode
                      2 : Auto mode
              
        channel     : returns the channel used to trigger
                      0 : chA
                      1 : chB
        slope       : returns whether to trigger on rising or falling slope
                      0 : falling
                      1 : rising
        level       : returns the value the signal must cross (in the directin 
                      of the slope specified) to trigger
        hysteresis  : is used to prevent noise from causing a false trigger.
                      For a rising edge trigger, the signal must pass below
                      (level - hysteresis) before a trigger level crossing is
                      detected
                      For a falling edge trigger, the signal must pass above
                      (level + hysteresis) before a trigger level crossing is
                      detected
        position    : returns the pre-sampling trigger position in seconds

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.TrigGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(30)
        
        mode        = self.NanonisTCP.hex_to_int16(response[0:2])
        channel     = self.NanonisTCP.hex_to_int16(response[2:4])
        slope       = self.NanonisTCP.hex_to_int16(response[4:6])
        level       = self.NanonisTCP.hex_to_float64(response[6:14])
        hysteresis  = self.NanonisTCP.hex_to_float64(response[14:22])
        position    = self.NanonisTCP.hex_to_float64(response[22:30])
        
        return [mode,channel,slope,level,hysteresis,position]
    
    def DataGet(self,dataToGet):
        """
        Returns the graph data from the Oscilloscope 2-Channels

        Parameters
        ----------
        dataToGet : 'Current' returns the currently displayed data. 'Next' 
                    waits for the next trigger to retrieve data. 'Wait' waits
                    two triggers
                    0 : Current
                    1 : Next
                    2 : Wait
        Returns
        -------
        t0  : timestamp of the first acquired point
        dt  : time distance between two acquired points
        chA : numpy array of chA data
        chB : numpy array of chB data

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Osci2T.DataGet', body_size=2)
        
        hex_rep += self.NanonisTCP.to_hex(dataToGet,2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        t0  = self.NanonisTCP.hex_to_float64(response[0:8])
        dt  = self.NanonisTCP.hex_to_float64(response[8:16])
        pts = self.NanonisTCP.hex_to_int32(response[16:20])
        
        idx = 20
        chA = []
        for n in range(pts):
            chA.append(self.NanonisTCP.hex_to_float64(response[idx:idx+8]))
            idx += 8
        
        chB = []
        for n in range(pts):
            chB.append(self.NanonisTCP.hex_to_float64(response[idx:idx+8]))
            idx += 8
            
        return [t0,dt,np.array(chA),np.array(chB)]
        


# -*- coding: utf-8 -*-
"""
Created on Tue Apr 26 12:24:35 2022

@author: Julian Ceddia
"""
"""
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!WARNING: RUNNING run_test() WILL CHANGE SETTINGS IN NANONIS. RUN IT ON A SIM!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
"""

from nanonisTCP import nanonisTCP
from nanonisTCP.Scan import Scan

"""
Set up the TCP connection and interface
"""
def run_test(TCP_IP='127.0.0.1', TCP_PORT=6501, make_plot=False):
    
    # Listening Port: see Nanonis File>Settings>TCP Programming Interface
    NTCP = nanonisTCP(TCP_IP, TCP_PORT)                                             # Nanonis TCP interface
    
    try:
        scan = Scan(NTCP)                                                               # Nanonis Scan Module
        
        """
        Frame Parameters Set/Get
        """
        scan.FrameSet(5e-9,-5e-9,30e-9,30e-9,15)
        x,y,w,h,angle = scan.FrameGet()
        print("Frame Parameters")
        print("Frame Position:  " + str(x) + " m," + str(y) + " m")
        print("Frame dimension: " + str(w) + " m x " + str(h) + " m")
        print("Frame angle:     " + str(angle) + " deg")
        print("----------------------------------------------------------------------")

        """
        Buffer Parameters Set/Get
        """
        scan.BufferSet(channel_indexes=[0,14],pixels=128,lines=128)
        num_channels,channel_indexes,pixels,lines = scan.BufferGet()
        print("Buffer Parameters")
        print("Buffer num_channels:    " + str(num_channels))
        print("Buffer channel_indexes: " + str(channel_indexes))
        print("Buffer pixels:          " + str(pixels))
        print("Buffer lines:           " + str(lines))
        print("----------------------------------------------------------------------")

        """
        Props Set/Get
        """
        scan.PropsSet(bouncy_scan=1,comment="testing123")
        continuous_scan,bouncy_scan,autosave,series_name,comment = scan.PropsGet()
        print("Props Parameters")
        print("Props continuous_scan: " + str(continuous_scan))
        print("Props bouncy_scan:     " + str(bouncy_scan))
        print("Props autosave:        " + str(autosave))
        print("Props series_name:     " + str(series_name))
        print("Props comment:         " + str(comment))
        print("----------------------------------------------------------------------")

        """
        Speed Set/Get
        """
        # Best practice here is to define at least (fwd_speed or fwd_line_time) + (corresponding bwd param or speed_ratio)
        # e.g. one of these...
        # scan.SpeedSet(fwd_speed=150e-9,bwd_speed=400e-9)                              # Expect speeds to be set accordingly and constant param = line speed
        # scan.SpeedSet(fwd_line_time=1,bwd_line_time=0.25)                             # Expect times to be set accordingly and constant param = line time
        # scan.SpeedSet(fwd_speed=150e-9,speed_ratio=2)                                 # Expect fwd_speed=150e-9 and bwd_speed=300e-9
        scan.SpeedSet(fwd_line_time=1,speed_ratio=2)                                    # Expect fwd_line_time=1s and bwd_line_time=0.5s

        fwd_speed,bwd_speed,fwd_line_time,bwd_line_time,const_param,speed_ratio = scan.SpeedGet()
        print("Scan Speed Parameters")
        print("Forward speed:      " + str(fwd_speed))
        print("Backward speed:     " + str(bwd_speed))
        print("Forward line time:  " + str(fwd_line_time))
        print("Backward line time: " + str(bwd_line_time))
        print("Constant Param:     " + ["Constant speed", "Constant time"][const_param])
        print("Speed ratio:        " + str(speed_ratio))
        print("----------------------------------------------------------------------")
        
        
        """
        Action
        """
        scan.Action("start","down")

        """
        Status Get
        """

        """
        Wait end of scan
        """
        timeout_status, file_path_size, file_path = scan.WaitEndOfScan()
        print("Wait end of scan")
        print("timeout_status: " + ["EOS","Timed out"][timeout_status])
        print("file_path: " + file_path)
        print("----------------------------------------------------------------------")
       
        """
        FrameDataGrab
        """
        channel_name,scan_data,scan_direction = scan.FrameDataGrab(14, 1)
        print("Frame Data")
        print("Channel name:   " + channel_name)
        print("Scan direction: " + scan_direction)
        print("Run commented code to show scan_data")
        if make_plot == True:
            import matplotlib.pyplot as plt
            import matplotlib
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.imshow(scan_data)
        print("----------------------------------------------------------------------")

    except:
        print('error')
    
    NTCP.close_connection()

    return print('end of test')


# -*- coding: utf-8 -*-
"""
Created on Wed May  4 13:52:22 2022

@author: ben

Updated by
@author Julian Ceddia 3/3/22
"""

class TipShaper:
    """
    Nanonis Tip Shaper Module
    """
    def __init__(self,NanonisTCP):
        self.NanonisTCP = NanonisTCP
        
    def Start(self,wait_until_finished,timeout):
        """
        Starts the tip shaper procedure.

        Parameters
        ----------
        wait_until_finished : defines if this function waits until Tip Shaper
                              procedure stops.
                              0 = False (does not wait)
                              1 = True (waits)
                              
        timeout             : sets the number of milliseconds to wait if 
                            wait_until_finished is set to True. A value of -1 
                            means waiting forever [ms]. 
        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('TipShaper.Start', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(wait_until_finished,4)
        hex_rep += self.NanonisTCP.to_hex(timeout,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
    def PropsSet(self,Switch_off_delay="-default",Change_bias="-default", Bias="-default", Tip_lift="-default", Lift_time_1="-default",
                 Bias_lift="-default", Bias_settling_time="-default", Lift_height="-default", Lift_time_2="-default", 
                 End_wait_time="-default", Restore_feedback="-default"):
        """
        Sets the configuration of the tip shaper procedure. All inputs are 
        optional... -default will leave the setting as is in Nanonis

        Parameters
        ----------
        Switch_off_delay    : the time during which the Z position is averaged before
                            switching the Z controller off.
                            
        Change_bias         : decides whether the Bias value is applied right
                            before first Z ramping.
                            0 = no change
                            1 = True
                            2 = False
                            
        Bias                : is the value applied to the Bias signal if 
                            Change_bias is True [V]. 
                            
        Tip_lift            : defines the relative height the tip is going to ramp
                            for the first time (from current Z position) [m].
                            
        Lift_time_1         : defines the time to ramp Z from current Z position
                            by the Tip_lift amount [s]. 
                            
        Bias_lift           : the Bias voltage applied just after the first Z
                            ramping [V]. 
                            
        Bias_settling_time  : the time to wait after applying the Bias Lift value,
                            and it is also the time to wait after applying Bias
                            before ramping Z for the first time [s]. 
                            
        Lift_height         : returns the height the tip is going to ramp for the
                            second time [m]. 
                            
        Lift_time_2         : given time to ramp Z in the second ramping [s].
                            
        End_wait_time       : time to wait after restoring the initial Bias voltage
                            (just after finishing second ramping) [s].
                            
        Restore_feedback    : returns whether the initial Z-controller status is
                            restored at the end of the tip shaper procedure.
                            0 = False
                            1 = True

        """
        default_args = self.PropsGet()                                          # Store all the current tip shaping settings in nanonis
        
        # Order matters here
        tipShaperArgs = [Switch_off_delay]
        tipShaperArgs.append(Change_bias)
        tipShaperArgs.append(Bias)
        tipShaperArgs.append(Tip_lift)
        tipShaperArgs.append(Lift_time_1)
        tipShaperArgs.append(Bias_lift)
        tipShaperArgs.append(Bias_settling_time)
        tipShaperArgs.append(Lift_height)
        tipShaperArgs.append(Lift_time_2)
        tipShaperArgs.append(End_wait_time)
        tipShaperArgs.append(Restore_feedback)
        
        for i,a in enumerate(tipShaperArgs):
            if(a=="-default"):
                tipShaperArgs[i] = default_args[i]
                
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('TipShaper.PropsSet', body_size=44)
        
        ## Arguments 
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[0])
        hex_rep += self.NanonisTCP.to_hex(tipShaperArgs[1],4)
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[2])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[3])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[4])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[5])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[6])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[7])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[8])
        hex_rep += self.NanonisTCP.float32_to_hex(tipShaperArgs[9])
        hex_rep += self.NanonisTCP.to_hex(tipShaperArgs[10],4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
    def PropsGet(self):
        """
        Returns the configuration of the tip shaper procedure.

        Returns
        -------
        Switch_off_delay    : the time during which the Z position is averaged before
                            switching the Z controller off.
                            
        Change_bias         : decides whether the Bias value is applied right
                            before first Z ramping.
                            0 = no change
                            1 = True
                            2 = False
                            
        Bias                : is the value applied to the Bias signal if 
                            Change_bias is True [V]. 
                            
        Tip_lift            : defines the relative height the tip is going to ramp
                            for the first time (from current Z position) [m].
                            
        Lift_time_1         : defines the time to ramp Z from current Z position
                            by the Tip_lift amount [s]. 
                            
        Bias_lift           : the Bias voltage applied just after the first Z
                            ramping [V]. 
                            
        Bias_settling_time  : the time to wait after applying the Bias Lift value,
                            and it is also the time to wait after applying Bias
                            before ramping Z for the first time [s]. 
                            
        Lift_height         : returns the height the tip is going to ramp for the
                            second time [m]. 
                            
        Lift_time_2         : given time to ramp Z in the second ramping [s].
                            
        End_wait_time       : time to wait after restoring the initial Bias voltage
                            (just after finishing second ramping) [s].
                            
        Restore_feedback    : returns whether the initial Z-controller status is
                            restored at the end of the tip shaper procedure.
                            0 = False
                            1 = True

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('TipShaper.PropsGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(44)
        
        Switch_off_delay    = self.NanonisTCP.hex_to_float32(response[0:4])
        Change_bias         = self.NanonisTCP.hex_to_uint32(response[4:8])
        Bias                = self.NanonisTCP.hex_to_float32(response[8:12])
        Tip_lift            = self.NanonisTCP.hex_to_float32(response[12:16])
        Lift_time_1         = self.NanonisTCP.hex_to_float32(response[16:20])
        Bias_lift           = self.NanonisTCP.hex_to_float32(response[20:24])
        Bias_settling_time  = self.NanonisTCP.hex_to_float32(response[24:28])
        Lift_height         = self.NanonisTCP.hex_to_float32(response[28:32])
        Lift_time_2         = self.NanonisTCP.hex_to_float32(response[32:36])
        End_wait_time       = self.NanonisTCP.hex_to_float32(response[36:40])
        Restore_feedback    = self.NanonisTCP.hex_to_uint32(response[40:44])
        
        return [Switch_off_delay,Change_bias,Bias,Tip_lift,Lift_time_1,Bias_lift,
                Bias_settling_time,Lift_height,Lift_time_2,End_wait_time,
                Restore_feedback]


# -*- coding: utf-8 -*-
"""
Created on Sat Apr 16 15:40:21 2022

@author: jced0001
"""

class FolMe:
    """
    Nanonis Follow Me Module
    """
    def __init__(self, nanonisTCP):
        self.nanonisTCP = nanonisTCP
    
    def XYPosSet(self, X, Y, Wait_end_of_move=False):
        """
        This function moves the tip to the specified X and Y target coordinates
        (in meters). It moves at the speed specified by the "Speed" parameter
        in the Follow Me mode of the Scan Control module. This function will 
        return when the tip reaches its destination or if the movement stops.

        Parameters
        X : Set x position (m)
        Y : Set y position (m)
        Wait_end_of_move : False: Selects whether the function  immediately
                           True: Waits until tip stops moving
        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.XYPosSet', body_size=20)
        
        ## arguments
        hex_rep += self.nanonisTCP.float64_to_hex(X)
        hex_rep += self.nanonisTCP.float64_to_hex(Y)
        hex_rep += self.nanonisTCP.to_hex(Wait_end_of_move,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        self.nanonisTCP.receive_response(0)

    def XYPosGet(self, Wait_for_newest_data):
        """
        Returns the X,Y tip coordinates

        Parameters
        Wait_for_newest_data (uint32): 

        Returns
        xpos : x position of the tip (m)
        ypos : y position of the tip (m)

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.XYPosGet', body_size=4)
        
        # arguments
        hex_rep += self.nanonisTCP.to_hex(Wait_for_newest_data,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        ## Receive Response
        response = self.nanonisTCP.receive_response(16)
        
        xpos = self.nanonisTCP.hex_to_float64(response[0:8])
        ypos = self.nanonisTCP.hex_to_float64(response[8:16])
        
        return (xpos,ypos)
    
    def SpeedSet(self, speed, custom_speed):
        """
        Configures the tip speed when moving in Follow Me

        Parameters
        speed (float32)       : sets the surface speed in Follow Me
        custom_speed (uint32) : True:  speed setting is custom speed
                                False: speed setting is scan speed
                                
        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.SpeedSet', body_size=8)
        
        # arguments
        hex_rep += self.nanonisTCP.float32_to_hex(speed)
        hex_rep += self.nanonisTCP.to_hex(custom_speed,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check for errors)
        self.nanonisTCP.receive_response(0)
    
    def SpeedGet(self):
        """
        Returns the tip speed when moving in Follow Me mode

        Returns
        -------
        speed (float32)       : surface speed in Follow Me mode
        custom_speed (uint32) : True:  speed setting is custom speed
                                False: speed setting is scam speed
                                
        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.SpeedGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response(8)
        
        speed        = self.nanonisTCP.hex_to_float32(response[0:4])
        custom_speed = self.nanonisTCP.hex_to_uint32(response[4:8]) > 0
        
        return (speed,custom_speed)
    
    def OversamplSet(self,Oversampling):
        """
        Sets the oversampling of the acquired data when the tip is moving in
        Follow Me mode

        Parameters
        Oversampling (int) : oversampling of data acquisition in fol me mode

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.OversamplSet', body_size=4)
        
        # arguments
        hex_rep += self.nanonisTCP.to_hex(Oversampling,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check for errors only)
        self.nanonisTCP.receive_response(0)
    
    def OversamplGet(self):
        """
        Returns the overampling and rate of the acquired data when tip is in
        Follow Me mode

        Returns
        oversampling (int)    : oversampling of data acquisition in fol me mode
        sample_rate (float32) : sample rate

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.OversamplGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response(8)
        
        oversampling = self.nanonisTCP.hex_to_int32(response[0:4])
        sample_rate  = self.nanonisTCP.hex_to_float32(response[4:8])
        
        return (oversampling, sample_rate)
    
    def Stop(self):
        """
        Stops the tip movement in follow me mode

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.Stop', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
    
    def PSOnOffSet(self,ps_status):
        """
        Enables or disables Point & Shoot in Follow Me mode

        Parameters
        ps_status (boolean) : True:  enabled
                              False: disabled

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.PSOnOffSet', body_size=4)
        
        # arguments
        hex_rep += self.nanonisTCP.to_hex(ps_status,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
        
    def PSOnOffGet(self):
        """
        Returns if Point & Shoot is enabled or disabled in follow me mode

        Returns
        ps_status (boolean): True:  enabled
                             False: disabled

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('FolMe.PSOnOffGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response(4)
        
        ps_status = self.nanonisTCP.hex_to_uint32(response[0:4]) > 0
        
        return ps_status
    


# -*- coding: utf-8 -*-
"""
Created on Thu Jun  9 22:51:52 2022

@author: jced0001
"""

class Piezo:
    """
    Nanonis Piezo Module
    """
    def __init__(self,NanonisTCP):
        self.NanonisTCP = NanonisTCP
    
    def TiltSet(self,tilt_x=None,tilt_y=None):
        """
        Configures the tilt correction parameters. Passing in None for either x
        or y tilt keeps the setting as is in nanonis.

        Parameters
        ----------
        tilt_x : Sets by which angle to correct the tilt in the X direction
        tilt_y : Sets by which angle to correct the tilt in the Y direction

        """
        n_tilt_x,n_tilt_y = self.TiltGet()
        if(not tilt_x): tilt_x = n_tilt_x
        if(not tilt_y): tilt_y = n_tilt_y
        
        hex_rep = self.NanonisTCP.make_header('Piezo.TiltSet', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(tilt_x)
        hex_rep += self.NanonisTCP.float32_to_hex(tilt_y)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def TiltGet(self):
        """
        Returns the tilt correction parameters.

        Returns
        -------
        tilt_x : Sets by which angle to correct the tilt in the X direction
        tilt_y : Sets by which angle to correct the tilt in the Y direction

        """
        hex_rep = self.NanonisTCP.make_header('Piezo.TiltGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(8)
        
        tilt_x = self.NanonisTCP.hex_to_float32(response[0:4])
        tilt_y = self.NanonisTCP.hex_to_float32(response[4:8])
        
        return [tilt_x,tilt_y]
    
    def RangeSet(self,range_x=None,range_y=None,range_z=None):
        """
        Sets the piezo range (m) values for all 3 aces (X,Y,Z). Leave param as 
        None means it will remain as is in nanonis.

        Parameters
        ----------
        range_x : range of the X piezo (m)
        range_y : range of the Y piezo (m)
        range_z : range of the Z piezo (m)

        """
        n_range_x,n_range_y,n_range_z = self.RangeGet()
        if(not range_x): range_x = n_range_x
        if(not range_y): range_y = n_range_y
        if(not range_z): range_z = n_range_z
    
        hex_rep = self.NanonisTCP.make_header('Piezo.RangeSet', body_size=12)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(range_x)
        hex_rep += self.NanonisTCP.float32_to_hex(range_y)
        hex_rep += self.NanonisTCP.float32_to_hex(range_z)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
    def RangeGet(self):
        """
        Returns the piezo range (m) for all 3 axes (X,Y,Z)

        Returns
        -------
        range_x : range of the X piezo (m)
        range_y : range of the Y piezo (m)
        range_z : range of the Z piezo (m)

        """
        hex_rep = self.NanonisTCP.make_header('Piezo.RangeGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(12)
        
        range_x = self.NanonisTCP.hex_to_float32(response[0:4])
        range_y = self.NanonisTCP.hex_to_float32(response[4:8])
        range_z = self.NanonisTCP.hex_to_float32(response[8:12])
        
        return [range_x,range_y,range_z]
    
    def DriftCompGet(self):
        """
        Returns the drift compensation parameters

        Returns
        ----------
        on    : True: Turn compensation on
                False: Turn compensation off
        vx    : linear speed applied to the X piezo (m/s)
        vy    : linear speed applied to the Y piezo (m/s)
        vz    : linear speed applied to the Z piezo (m/s)
        xsat  : indicates if the X drift correction reached 10% of piezo range
        ysat  : indicates if the Y drift correction reached 10% of piezo range
        zsat  : indicates if the Z drift correction reached 10% of piezo range

        """
        hex_rep = self.NanonisTCP.make_header('Piezo.DriftCompGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(28)
        
        status = self.NanonisTCP.hex_to_uint32(response[0:4])
        vx     = self.NanonisTCP.hex_to_float32(response[4:8])
        vy     = self.NanonisTCP.hex_to_float32(response[8:12])
        vz     = self.NanonisTCP.hex_to_float32(response[12:16])
        xsat   = self.NanonisTCP.hex_to_uint32(response[16:20])
        ysat   = self.NanonisTCP.hex_to_uint32(response[20:24])
        zsat   = self.NanonisTCP.hex_to_uint32(response[24:28])
        
        return [status,vx,vy,vz,xsat,ysat,zsat]
        
    def DriftCompSet(self,on,vx=[],vy=[],vz=[]):
        """
        Configures the drift compensation parameters

        Parameters
        ----------
        on : True: Turn compensation on
             False: Turn compensation off
        vx : linear speed applied to the X piezo (m/s)
        vy : linear speed applied to the Y piezo (m/s)
        vz : linear speed applied to the Z piezo (m/s)

        """
        _,n_vx,n_vy,n_vz,_,_,_ = self.DriftCompGet()
        if(type(vx) == list): vx = n_vx
        if(type(vy) == list): vy = n_vy
        if(type(vz) == list): vz = n_vz
    
        hex_rep = self.NanonisTCP.make_header('Piezo.DriftCompSet', body_size=16)
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(on,4)
        hex_rep += self.NanonisTCP.float32_to_hex(vx)
        hex_rep += self.NanonisTCP.float32_to_hex(vy)
        hex_rep += self.NanonisTCP.float32_to_hex(vz)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Jun  9 23:09:19 2022

@author: jced0001
"""
"""
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!WARNING: RUNNING run_test() WILL CHANGE SETTINGS IN NANONIS. RUN IT ON A SIM!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
"""

from nanonisTCP import nanonisTCP
# from Piezo import Piezo
from nanonisTCP.Piezo import Piezo

"""
Set up the TCP connection and interface
"""
def run_test(TCP_IP='127.0.0.1', TCP_PORT=6501):
    # Listening Port: see Nanonis File>Settings>TCP Programming Interface
    NTCP = nanonisTCP(TCP_IP, TCP_PORT)                                         # Nanonis TCP interface
    try:
        piezo = Piezo(NTCP)                                                     # Nanonis TCP Piezo Module
        
        """
        Tilt Set/Get
        """
        piezo.TiltSet(0.1,-0.2)
        tilt_x,tilt_y = piezo.TiltGet()
        print("Piezo Tilt")
        print("Tilt x: " + str(tilt_x))
        print("Tilt y: " + str(tilt_y))
        print("----------------------------------------------------------------------")
        
        """
        Range Set/Get
        """
        piezo.RangeSet(1e-6,1e-6,360e-9)
        range_x,range_y,range_z = piezo.RangeGet()
        print("Piezo Ranges")
        print("Range x: " + str(range_x) + " m")
        print("Range y: " + str(range_y) + " m")
        print("Range z: " + str(range_z) + " m")
        print("----------------------------------------------------------------------")
        
        
        """
        Compensation Set/Get
        """
        piezo.DriftCompSet(on=False,vx=1e-15,vy=2e-15,vz=3e-15)
        status,vx,vy,vz,xstat,ystat,zstat = piezo.DriftCompGet()
        print("Piezo drift compensation")
        print("Status: " + str(status))
        print("X: " + str(vx) + " m/s")
        print("Y: " + str(vy) + " m/s")
        print("Z: " + str(vz) + " m/s")
        print("----------------------------------------------------------------------")
        
    except Exception as e:
        print(str(e))
    NTCP.close_connection()
    return print('end of test')




#!/usr/bin/env python

from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

version = '1.0.2'

setup(
    name='nanonisTCP',
    version=version,
    author='Julian Ceddia',
    author_email='jdceddia@gmail.com',
    description='python module for communicating via nanonis TCP protocal',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/New-Horizons-SPM/nanonisTCP',
    project_urls = {
        "Bug Tracker": "https://github.com/New-Horizons-SPM/nanonisTCP/issues"
    },
    license='MIT',
    packages=find_packages(),
    install_requires=['numpy', 'matplotlib'],
)



# -*- coding: utf-8 -*-
"""
Created on Tue May  3 15:13:02 2022

@author: jced0001
"""

class SafeTip:
    """
    Nanonis SafeTip Module
    """
    def __init__(self, NanonisTCP):
        self.NanonisTCP = NanonisTCP
    
    def OnOffSet(self,safe_tip_status):
        """
        Switches the Safe Tip feature on or off

        Parameters
        ----------
        safe_tip_status : sets the Safe Tip feature on or off
                          0: No change (leave setting as is in nanonis)
                          1: Safe Tip feature on
                          2: Safe Tip feature off

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('SafeTip.OnOffSet', body_size=2)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(safe_tip_status,2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def OnOffGet(self):
        """
        Returns the Safe Tip feature status (on/off)

        Returns
        -------
        safe_tip_status : sets the Safe Tip feature on or off
                          True:  Safe Tip feature on
                          False: Safe Tip feature off

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('SafeTip.OnOffGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(2)
        
        safe_tip_status = self.NanonisTCP.hex_to_uint16(response[0:2])
        
        return safe_tip_status
    
    def SignalGet(self):
        """
        Returns the current Safe Tip signal value

        Returns
        -------
        signal_value : signal value

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('SafeTip.SignalGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        signal_value = self.NanonisTCP.hex_to_float32(response[0:4])
        
        return signal_value
    
    def PropsSet(self,auto_recovery,auto_pause_scan,threshold):
        """
        Sets the safe tip configuration

        Parameters
        ----------
        auto_recovery   : indicates if Z-Controller automatically recovers from
                          a SafeTip situation after a specified amount of time 
                          if Z-Controller was originall on.
                          False: Off
                          True:  On
        auto_pause_scan : indicates if the Z-Controller automatically pauses/
                          holds the scan on a SafeTip event.
                          False: Off
                          True:  On
        threshold       : defines the condition to trigger the Safe Tip
        
        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('SafeTip.PropsSet', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(auto_recovery,2)
        hex_rep += self.NanonisTCP.to_hex(auto_pause_scan,2)
        hex_rep += self.NanonisTCP.float32_to_hex(threshold)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def PropsGet(self):
        """
        Returns the Safe Tip configuration

        Returns
        -------
        auto_recovery   : indicates if Z-Controller automatically recovers from
                          a SafeTip situation after a specified amount of time 
                          if Z-Controller was originall on.
                          False: Off
                          True:  On
        auto_pause_scan : indicates if the Z-Controller automatically pauses/
                          holds the scan on a SafeTip event.
                          False: Off
                          True:  On
        threshold       : defines the condition to trigger the Safe Tip

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('SafeTip.PropsGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(8)
        
        auto_recovery   = self.NanonisTCP.hex_to_uint16(response[0:2])
        auto_pause_scan = self.NanonisTCP.hex_to_uint16(response[2:4])
        threshold       = self.NanonisTCP.hex_to_float32(response[4:8])
        
        return [auto_recovery,auto_pause_scan,threshold]


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 9 17:15:00 2023

@author: Liam
"""

import numpy as np

class Pattern:
    """
    Nanonis Pattern Module
    """
    def __init__(self,NanonisTCP):
        self.NanonisTCP = NanonisTCP
        
    def ExpOpen(self):
        """
        Opens the selected grid experiment.

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Pattern.ExpOpen', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def ExpStart(self,pattern):
        """
        Starts the selected grid experiment.

        Before using this function, select the experiment through Pattern.PropsSet, 
        and be sure to have it open in the software or through the function 
        Pattern.ExpOpen. Otherwise it will give an error saying that the experiment 
        has not been configured yet.

        Parameters
        ----------
        pattern     : defines switches the active pattern to this value before 
                      starting the grid experiment. 
                        : 0 = no change
                        : 1 = Grid
                        : 2 = Line
                        : 3 = Cloud
                    

        Returns
        -------
        None.

        """
        # Create int assignments to allow string inputs
        pattern_dict = {
            'no change' : 0,
            'Grid'      : 1,
            'Line'      : 2,
            'Cloud'     : 3
        }

        # Make sure input is one of the strings or int assignments
        if pattern not in pattern_dict.keys() and pattern not in pattern_dict.values():
            raise Exception('pattern input must be either:\n0 = no change \n1 = Grid\n2 = Line\n3 = Cloud')
        # Convert string to int assignment
        if type(pattern) is str: pattern = pattern_dict[pattern]

        body_size = 2                                                       # 2 bytes for pattern (uint16)
        
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Pattern.ExpStart', body_size=body_size)
        
        ## arguments
        hex_rep += self.NanonisTCP.to_hex(pattern,2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive Response
        self.NanonisTCP.receive_response(0)

    def ExpPause(self, pause):
        """
        Pauses or resumes the selected grid experiment.

        Parameters
        ----------
        pause:   : 0 = Resume
                 : 1 = Pause

        Issues
        ------
        Cant pause on the same socket that the experiment was started from.
        If you want to pause or stop, open a new TCP connection under a different port 
        or disconnect and reconnect with the current port.

        """

        pause_dict = {
            'resume': 0,
            'pause' : 1
        }

        if type(pause) is str: pause = pause_dict[pause]

        body_size = 4                                                       # 2 bytes for pause (uint32)
        
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Pattern.ExpPause', body_size=body_size)
        
        ## arguments
        hex_rep += self.NanonisTCP.to_hex(pause,4)

        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)

    
    def ExpStop(self):
        """
        Stops the selected grid experiment.

        Issues
        ------
        Cant pause on the same socket that the experiment was started from.
        If you want to pause or stop, open a new TCP connection under a different port 
        or disconnect and reconnect with the current port.

        """

        hex_rep = self.NanonisTCP.make_header('Pattern.ExpStop', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def ExpStatusGet(self):
        """
        Returns the status of the selected grid experiment.
        
        Returns
        -------
        status : 0: Not running. 1: Running

        """
        hex_rep = self.NanonisTCP.make_header('Pattern.ExpStatusGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        status = self.NanonisTCP.hex_to_int32(response[0:4])
        
        return status
        
    def GridSet(self,num_points_x,num_points_y,x,y,w,h,angle):
        """
        Sets the grid size parameters.

        Parameters
        ----------
        num_points_x            : number of points in x that defines the grid
        num_points_y            : number of points in y that defines the grid
        x                       : x (m) coordinate of the centre of the grid 
        y                       : y (m) coordinate of the centre of the grid 
        w                       : the width (m) of the grid 
        h                       : the height (m) of the grid 
        angle                   : the rotation angle (deg) of the grid

        Returns
        -------
        None.

        Defaults
        --------
        set_active_pattern      : defines if the pattern switches to Grid, 
                                  in case it was not Cloud already.
                                    : 0 = Off 
                                    : 1 = On

                                  DEFAULT = 1

        grid_scan_frame         : defines if the grid size should be set like 
                                  the scan frame size. 
                                  : 0 = No 
                                  : 1 = Yes

                                DEFAULT = 0

        """
            
        body_size = 4 + 2*4 + 4 + 5*4                                    # 4 bytes for set_active_pattern (uint32), 2 x 4 bytes for num_points_x+y (int), 4 bytes for grid_scan_frame (uint32), 5 x 4 bytes for x, y, w, h, angle (float32)
        
        hex_rep = self.NanonisTCP.make_header('Pattern.GridSet', body_size=body_size)

        ## arguments

        set_active_pattern  = 1                                         # Useless, setting default to 1. Why call Grid set if not to activate the Grid?
        grid_scan_frame     = 0                                         # Useless, this function is helpful in the Nanonis GUI, but not in command line.

        hex_rep += self.NanonisTCP.to_hex(set_active_pattern,4)         
        hex_rep += self.NanonisTCP.to_hex(num_points_x,4)
        hex_rep += self.NanonisTCP.to_hex(num_points_y,4)
        hex_rep += self.NanonisTCP.to_hex(grid_scan_frame,4)            
        hex_rep += self.NanonisTCP.float32_to_hex(x)
        hex_rep += self.NanonisTCP.float32_to_hex(y)
        hex_rep += self.NanonisTCP.float32_to_hex(w)
        hex_rep += self.NanonisTCP.float32_to_hex(h)
        hex_rep += self.NanonisTCP.float32_to_hex(angle)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
    def GridGet(self):
        """
        Returns the grid size parameters.

        Returns
        -------
        num_points_x   : number of points in x that defines the grid
        num_points_y   : number of points in y that defines the grid
        x              : x (m) coordinate of the centre of the grid 
        y              : y (m) coordinate of the centre of the grid 
        w              : the width (m) of the grid 
        h              : the height (m) of the grid 
        angle          : the rotation angle (deg) of the grid


        """
        hex_rep = self.NanonisTCP.make_header('Pattern.GridGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        idx = 0
        args = []
        for i in range(2):                                              # Two int32 responses requiring 4 bytes each
            args.append(self.NanonisTCP.hex_to_int32(response[idx:idx+4]))
            idx += 4
        
        for i in range(5):                                              # Five float32 responses requiring 4 bytes each
            args.append(self.NanonisTCP.hex_to_float32(response[idx:idx+4]))
            idx += 4

        num_points_x, num_points_y, x, y, w, h, angle = args
        
        return [num_points_x, num_points_y, x, y, w, h, angle]
    
    def LineSet(self,num_points,x1,y1,x2,y2):
        """
        Sets the line size parameters.

        Parameters
        ----------
        num_points    : number of points that defines the line
        x1            : x (m) coordinate of the start of the line 
        y1            : y (m) coordinate of the start of the line 
        x2            : x (m) coordinate of the end of the line 
        y2            : y (m) coordinate of the end of the line 

        Returns
        -------
        None.

        Default Values
        --------------
        set_active_pattern      : defines if the pattern switches to Line, 
                                  in case it was not Line already.
                                    : 0 = Off 
                                    : 1 = On

                                  DEFAULT = 1

        line_scan_frame         : defines if the line size should be set like 
                                  the scan frame size. 
                                  : 0 = No 
                                  : 1 = Yes

                                DEFAULT = 0

        """
            
        body_size = 4 + 4 + 4 + 4*4                                    # 4 bytes for set_active_pattern (uint32), 4 bytes for num_points (int), 4 bytes for line_scan_frame (uint32), 4 x 4 bytes for x1, y1, x2, y2 (float32)
        
        hex_rep = self.NanonisTCP.make_header('Pattern.LineSet', body_size=body_size)

        ## arguments

        set_active_pattern  = 1                                            # Useless, setting default to 1. Why call Grid set if not to activate the Grid?
        line_scan_frame     = 0                                            # Useless, this function is helpful in the Nanonis GUI, but not in command line.

        hex_rep += self.NanonisTCP.to_hex(set_active_pattern,4)         
        hex_rep += self.NanonisTCP.to_hex(num_points,4)
        hex_rep += self.NanonisTCP.to_hex(line_scan_frame,4)            
        hex_rep += self.NanonisTCP.float32_to_hex(x1)
        hex_rep += self.NanonisTCP.float32_to_hex(y1)
        hex_rep += self.NanonisTCP.float32_to_hex(x2)
        hex_rep += self.NanonisTCP.float32_to_hex(y2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)

    def LineGet(self):
        """
        Returns the line size parameters.

        Returns
        -------
        num_points     : number of points that defines the line
        x1             : x (m) coordinate of the start of the line 
        y1             : y (m) coordinate of the start of the line 
        x2             : x (m) coordinate of the end of the line 
        y2             : y (m) coordinate of the end of the line 


        """
        hex_rep = self.NanonisTCP.make_header('Pattern.LineGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        
        args = []
        args.append(self.NanonisTCP.hex_to_int32(response[0:4]))
        
        idx = 4
        for i in range(4):                                              # Four float32 responses requiring 4 bytes each
            args.append(self.NanonisTCP.hex_to_float32(response[idx:idx+4]))
            idx += 4

        num_points, x1, y1, x2, y2 = args
        
        return [num_points, x1, y1, x2, y2]

    def CloudSet(self,x,y):
        """
        Configures a cloud of points.

        Parameters
        ----------
        x   : x (m) is a 1D array of the x coordinates of the of the points defining the cloud
        y   : y (m) is a 1D array of the y coordinates of the of the points defining the cloud

        Returns
        -------
        None.


        Default Values
        --------------
        set_active_pattern      : defines if the pattern switches to Cloud, 
                                  in case it was not Cloud already.
                                  0 = Off, 1 = On
                                 
                                  DEFAULT = 1

        num_points              : number of points that defines the cloud, AND
                                  it defines the size of the 1D arrays for x
                                  and y coordinates
                                  
                                  DEFAULT = len(x)

        """

        if not type(x) is list: 
            raise Exception('x must be a 1D list')
        
        if not type(y) is list: 
            raise Exception('y must be a 1D list')

        if np.array(x).ndim != 1 or np.array(y).ndim != 1:
            raise Exception('x and y inputs must be separate 1D lists of ordinates')
        
        if len(x) != len(y):
            raise Exception('x and y inputs must have the same number of elements (x (%.0i), and y (%.0i))' % (len(x), len(y)))

        set_active_pattern=1                                                      # Useless, setting default to 1. Why call Cloud set if not to activate the Cloud?
        
        num_points=len(x)                                                         # Calculate num_points from length of input vector.
        
        body_size = 4 + 4 + num_points*(4 + 4)                                    # 4 bytes for set_active_pattern (uint32), 4 bytes for num_points (int), 2 * 4 * num_points bytes for the total number of elements in both x and y 1D arrays (float32)
        
        hex_rep = self.NanonisTCP.make_header('Pattern.CloudSet', body_size=body_size)

        ## arguments

        hex_rep += self.NanonisTCP.to_hex(set_active_pattern,4)         
        
        hex_rep += self.NanonisTCP.to_hex(num_points,4)

        for p in x:
            hex_rep += self.NanonisTCP.float32_to_hex(p)
        
        for p in y:
            hex_rep += self.NanonisTCP.float32_to_hex(p)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)

    def CloudGet(self):
        """
        Returns the cloud configuration.

        Returns
        -------
        num_points     : number of points that defines the cloud
        x              : x (m) coordinates of the cloud 
        y              : y (m) coordinates of the cloud

        """
        hex_rep = self.NanonisTCP.make_header('Pattern.CloudGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        
        num_points = self.NanonisTCP.hex_to_int32(response[0:4])
        
        idx = 4
        x = []
        for i in range(num_points):                                              # Four bytes for each point in x (float32)
            x.append(self.NanonisTCP.hex_to_float32(response[idx:idx+4]))
            idx += 4

        y = []
        for i in range(num_points):                                              # Four bytes for each point in y (float32)
            y.append(self.NanonisTCP.hex_to_float32(response[idx:idx+4]))
            idx += 4
        
        return [num_points, x, y]

        
    def PropsSet(self,exp_name,basename,premeasure_delay,ext_VI_path='/',save_scan_channels=1):
        """
        Sets the configuration of the Grid experiment section in the Scan module.

        Returns
        -------
        exp_name            : is the name of the selected experiment. 
                              Must be one of: (example)
                                : 'Bias Spectroscopy ASCII'
                                : 'Bias Spectroscopy BINARY'
                                : 'Bias Sweep'
                                : 'Generic Sweep'
                                : 'Z Spectroscopy ASCII'
                                : 'Z Spectroscopy BINARY'
                                : 'Approach-Retract'
                                : 'DIO Trigger'
                                : 'External VI...'
                              For a full list of experiment names, see Pattern.PropsGet()
        basename            : sets the basename of the experiment file
        premeasure_delay    : is the time (s) to wait on each point before 
                              performing the experiment 
        ext_VI_path         : sets the path of the external VI. Relevant only for
                              exp_name = 'External VI...'
        save_scan_channels  : sets if the scan channels are saved into the grid 
                              experiment file. 
                                : 0 = Off 
                                : 1 = On

        """

        exp_names = ['Bias Spectroscopy ASCII', 'Bias Spectroscopy BINARY', 'Bias Sweep', 'Generic Sweep', 'Z Spectroscopy ASCII', 'Z Spectroscopy BINARY', 'Approach-Retract', 'DIO Trigger', 'External VI...']

        if exp_name not in exp_names:
            raise Exception('Experiment name not valid. Check spelling and case (exact). Must be one of: ' + ', '.join(map(str,exp_names)))

        ###################
        #------NOTE-------#
        ###################
        # Choosing exp_name='External VI...' given weird result in Nanonis GUI
        # I think it is trying to find the name of the VI given the path,
        # but if the path is false it returns some hex code instead.
        # Need to test with appropriate external VI.
            # Also, basename will not update on the GUI but does actually affect the name of the saved files

        if exp_name == 'External VI...':
            print('Make sure you have passed the External VI path in PropsSet(ext_VI_path='')')

        body_size = 3*4 + int(len(self.NanonisTCP.string_to_hex(exp_name))/2) + int(len(self.NanonisTCP.string_to_hex(basename))/2) + int(len(self.NanonisTCP.string_to_hex(ext_VI_path))/2) + 4 + 4            # 4*3 bytes for each string length, 1 byte for each character in the strings (exp_name, basename, ext_VI_path), 4 bytes for pre-measure delay (float32), and 4 bytes for save_scan_channels (uint32).

        hex_rep = self.NanonisTCP.make_header('Pattern.PropsSet', body_size=body_size)
        
        ## arguments
        hex_rep += self.NanonisTCP.to_hex(len(exp_name),4)
        hex_rep += self.NanonisTCP.string_to_hex(exp_name)
        hex_rep += self.NanonisTCP.to_hex(len(basename),4)
        hex_rep += self.NanonisTCP.string_to_hex(basename)
        hex_rep += self.NanonisTCP.to_hex(len(ext_VI_path),4)
        hex_rep += self.NanonisTCP.string_to_hex(ext_VI_path)
        hex_rep += self.NanonisTCP.float32_to_hex(premeasure_delay)
        hex_rep += self.NanonisTCP.to_hex(save_scan_channels,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
    def PropsGet(self):
        """
        Returns the configuration of the Grid experiment section in the Scan module.

        Returns
        -------
        list_of_exp             : list of experiments available in the section "Pattern".
        exp                     : name of the selected experiment.
        ext_VI_path             : path of the external VI
        premeasure_delay        : the time (s) to wait on each point before performing the experiment.
        save_scan_channels      : indicates if the scan channels are saved into the grid experiment file.
                                :   0: Off
                                :   1: On

        """
        hex_rep = self.NanonisTCP.make_header('Pattern.PropsGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        # size_list_of_exp = self.NanonisTCP.hex_to_int32(response[0:4])        # Useless
        num_exp = self.NanonisTCP.hex_to_int32(response[4:8])
        # Loop over all experiment names
        list_of_exp = []
        idx = 8
        for i in range(num_exp):
            listed_exp_size = self.NanonisTCP.hex_to_int32(response[idx:idx+4])
            idx += 4
            list_of_exp.append(response[idx:idx+listed_exp_size].decode())
            idx += listed_exp_size
        
        exp_size = self.NanonisTCP.hex_to_int32(response[idx:idx+4])
        idx += 4
        exp = response[idx:idx+exp_size].decode()
        idx += exp_size

        ext_VI_path_size = self.NanonisTCP.hex_to_int32(response[idx:idx+4])
        idx += 4
        ext_VI_path = response[idx:idx+ext_VI_path_size].decode()
        idx += ext_VI_path_size

        premeasure_delay = self.NanonisTCP.hex_to_float32(response[idx:idx+4])
        idx += 4
        save_scan_channels = self.NanonisTCP.hex_to_int32(response[idx:idx+4])
        
        return [list_of_exp, exp, ext_VI_path, premeasure_delay, save_scan_channels]


# -*- coding: utf-8 -*-
"""
Created on Mon May  2 15:32:25 2022

@author: Julian Ceddia
"""

class ZController:
    """
    Nanonis Z-Controller Module
    """
    def __init__(self, NanonisTCP):
        self.NanonisTCP = NanonisTCP
    
    def ZPosSet(self,zpos):
        """
        Sets the Z position of the tip

        Parameters
        ----------
        zpos : Z position (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.ZPosSet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(zpos)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
        
    def ZPosGet(self):
        """
        Returns the current Z position of the tip

        Returns
        -------
        zpos : the current z position of the tip

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.ZPosGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        zpos = self.NanonisTCP.hex_to_float32(response[0:4])
        
        return zpos
    
    def OnOffSet(self,on):
        """
        Switches the Z-Controller On or Off

        Parameters
        ----------
        on : True:  turn controller on
             False: turn controller off

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.OnOffSet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(on,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def OnOffGet(self):
        """
        Returns the status of the Z-Controller
        
        Returns
        -------
        z_status : indicates if the controller is on or off
                   True:  Z-Controller is on
                   False: Z-Controller is off
                   
        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.OnOffGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        z_status = self.NanonisTCP.hex_to_uint32(response[0:4])
        
        return z_status
    
    def SetpntSet(self,setpoint):
        """
        Sets the stpoint of the Z-Controller

        Parameters
        ----------
        setpoint : setpoint of the z-controller (A)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.SetpntSet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(setpoint)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def SetpntGet(self):
        """
        Returns the setpoint current of the Z-Controller

        Returns
        -------
        setpoint : setpoint current of the z-controller (A)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.SetpntGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        setpoint = self.NanonisTCP.hex_to_float32(response[0:4])
        
        return setpoint
    
    def GainSet(self,p_gain,time_constant):
        """
        Sets the Z-Controller gains (P,I) and time settings

        Parameters
        ----------
        p_gain        : proportional gain of the regulation loop
        time_constant : time constant of the regulation loop
        i_gain        : integral gain of the regulation loop (I=P/T) (not used)

        """
        i_gain = p_gain/time_constant
        
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.GainSet', body_size=12)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(p_gain)
        hex_rep += self.NanonisTCP.float32_to_hex(time_constant)
        hex_rep += self.NanonisTCP.float32_to_hex(i_gain)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def GainGet(self):
        """
        Returns the Z-Controller gains (P,I) and time constant

        Returns
        -------
        None.

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.GainGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        p_gain        = self.NanonisTCP.hex_to_float32(response[0:4])
        time_constant = self.NanonisTCP.hex_to_float32(response[4:8])
        i_gain        = self.NanonisTCP.hex_to_float32(response[8:12])
        
        return [p_gain,time_constant,i_gain]
    
    def SwitchOffDelaySet(self,delay):
        """
        Sets the switch off delay in seconds of the Z-Controller

        Parameters
        ----------
        delay : Z-Controller switch off delay time (s)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.SwitchOffDelaySet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(delay)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def SwitchOffDelayGet(self):
        """
        Returns the switch off delay in seconds of the Z-Controller

        Returns
        -------
        delay : Z-Controller switch off delay time (s)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.SwitchOffDelayGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        delay = self.NanonisTCP.hex_to_float32(response[0:4])
        
        return delay
    
    def TipLiftSet(self,tip_lift):
        """
        Sets the TipLift of the Z-Controller

        Parameters
        ----------
        tip_lift : tip lift (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.TipLiftSet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(tip_lift)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def TipLiftGet(self):
        """
        Returns the TipLift of the Z-Controller

        Returns
        -------
        tip_lift : tip lift (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.TipLiftGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        tip_lift = self.NanonisTCP.hex_to_float32(response[0:4])
        
        return tip_lift
    
    def Home(self):
        """
        Moves the tip to its home position.
        
        This function moves the tip to the home position defined by the
        Home Absolute (m)/Home Relative (m) value. (Absolute and relative can 
        be switched in the controller configuration panel in the software or
        ZCtrl.HomePropsSet)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.Home', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
    
    def HomePropsSet(self,rel_abs,home_pos):
        """
        Sets the current status of the Z-Controller Home switch and its 
        corresponding position

        Parameters
        ----------
        rel_abs  : relative or absolute.
                   0: no change (leave setting as is)
                   1: home position is absolute
                   2: home is relative to the current position
        home_pos : the home position value (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.HomePropsSet', body_size=6)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(rel_abs,2)
        hex_rep += self.NanonisTCP.float32_to_hex(home_pos)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
    
    def HomePropsGet(self):
        """
        Returns the current status of the Z-Controller

        Returns
        -------
        rel_abs  : relative or absolute mode
                   False: home position is absolute
                   True:  home position is relative to current position
        home_pos : home position value (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.HomePropsGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(6)
        
        rel_abs  = self.NanonisTCP.hex_to_uint16(response[0:2])
        home_pos = self.NanonisTCP.hex_to_float32(response[2:6])
        
        return [rel_abs,home_pos]
    
    def ActiveCtrlSet(self,index):
        """
        Sets the active Z-Controller

        Parameters
        ----------
        index : the index out of the list of controllers which can be retrieved
                by ZCtrl.ControllersListGet

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.ActiveCtrlSet', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(index,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
        
    def CtrlListGet(self):
        """
        Returns the list of Z-Controllers and the index of the active 
        controller

        Returns
        -------
        controllers             : list of available Z-Controllers
        active_controller_index : index of the active Z-Controller

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.CtrlListGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response()
        
        # list_of_controllers_size = self.NanonisTCP.hex_to_int16(response[0:4]) # Not needed
        number_of_controllers = self.NanonisTCP.hex_to_int16(response[4:8])
        
        if(not number_of_controllers): return [[],-1]                           # Return an empty list and -1 for the index if there are no available controllers
        
        idx = 8
        controllers = []
        for i in range(number_of_controllers):
            size = self.NanonisTCP.hex_to_int16(response[idx:idx+4])
            idx += 4
            controllers.append(response[idx:idx+size].decode())
            idx += size
        
        active_controller_index = self.NanonisTCP.hex_to_int32(response[idx:idx+4])
        
        return [controllers,active_controller_index]
    
    def Withdraw(self, wait_until_finished, timeout=-1):
        """
        Withdraws the tip

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.Withdraw', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(wait_until_finished,4)
        hex_rep += self.NanonisTCP.to_hex(timeout,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
    
    def WithdrawRateSet(self,slew_rate):
        """
        Sets the Z-Controller withdraw slew rate

        Parameters
        ----------
        slew_rate : withdraw slew rate (m/s)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.WithdrawRateSet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(slew_rate)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
    
    def WithdrawRateGet(self):
        """
        Returns the withdraw slew rate in m/s

        Returns
        -------
        slew_rate : withdraw slew rate (m/s)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.WithdrawRateGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        slew_rate  = self.NanonisTCP.hex_to_float32(response[0:4])
        
        return slew_rate
    
    def LimitsEnabledSet(self,limit_z_status):
        """
        Enables or disables the Z position limits

        Parameters
        ----------
        limit_z_status : enables/disables the Z limits
                         True:  Z limits enabled
                         False: Z limits disabled

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.LimitsEnabledSet', body_size=4)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(limit_z_status,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
    
    def LimitsEnabledGet(self):
        """
        Returns if the Z limits are enabled or disabled

        Returns
        -------
        limit_z_status : enables/disables the Z limits
                         True:  Z limits enabled
                         False: Z limits disabled

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.LimitsEnabledGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(4)
        
        limit_z_status = self.NanonisTCP.hex_to_int32(response[0:4])
        
        return limit_z_status
    
    def LimitsSet(self,z_high_limit,z_low_limit):
        """
        Sets the Z position high and low limits in meters.
        
        When the Z position limits are not enabled, this function has no effect

        Parameters
        ----------
        z_high_limit : high limit (m)
        z_low_limit  : low limit (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.LimitsSet', body_size=8)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(z_high_limit)
        hex_rep += self.NanonisTCP.float32_to_hex(z_low_limit)
        
        self.NanonisTCP.send_command(hex_rep)
        
        # Receive response (check errors)
        self.NanonisTCP.receive_response(0)
    
    def LimitsGet(self):
        """
        Returns the Z position high and low limits in meters.
        
        When the Z position limits are not enabled, they correspond to the 
        piezo range limits.

        Returns
        -------
        z_high_limit : high limit (m) (if not enabled, then piezo high limit)
        z_low_limit  : low limit (m) (if not enabled, then piezo low limit)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.LimitsGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(8)
        
        z_high_limit = self.NanonisTCP.hex_to_float32(response[0:4])
        z_low_limit  = self.NanonisTCP.hex_to_float32(response[4:8])
        
        return [z_high_limit,z_low_limit]
    
    def StatusGet(self):
        """
        Returns the current status of the Z-Controller module

        Returns
        -------
        z_controller_status : returns if the controller is:
                              1: Off
                              2: On
                              3: Hold
                              4: Switching Off
                              5: Safe tip event occurred
                              6: Tip is currently withdrawing
        status_string       : String corresponding to the status
        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('ZCtrl.StatusGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(2)
        
        z_controller_status = self.NanonisTCP.hex_to_uint16(response[0:2])
        
        status_string = ["Off","On","Hold","Switching Off","Safe Tip Event","Withdrawing"][z_controller_status-1]
        
        return [z_controller_status,status_string]
        


# -*- coding: utf-8 -*-
"""
Created on Tue May  3 19:26:25 2022

@author: Julian Ceddia
"""

class Motor:
    """
    Nanonis Coarse Motion (Motor) Module
    """
    def __init__(self, NanonisTCP):
        self.NanonisTCP = NanonisTCP
    
    def StartMove(self,direction,steps,wait_until_finished=True,group=1):
        """
        Moves the coarse positioning device (motor, piezo actuator...)

        Parameters
        ----------
        direction : selects in which direction to move. Note that depending on 
                    your motor controller and setuo only the Z axis or Z- axis
                    may work
                    Possible values: "X+","X-","Y+","Y-","Z+","Z-"
        steps     : number of motor steps to move in the specified direction
        wait_until_finished : defines if this function only returns when the
                              motor reaches its destination or the movement
                              stops
                              False: Return immediately
                              True:  Return once function has reached its 
                                     destination
        group               : is the selection of groups defined in the motor
                              control module. If the motor doesn't support the
                              selection of groups, set it to 0
                              1: Group 1
                              2: Group 2
                              3: Group 3
                              4: Group 4
                              5: Group 5
                              6: Group 6

        """
        direction_dict = { "X+" : 0,
                           "X-" : 1,
                           "Y+" : 2,
                           "Y-" : 3,
                           "Z+" : 4,
                           "Z-" : 5,
                           }
        group -= 1                                                              # Make this more sensible
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.StartMove', body_size=14)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(direction_dict[direction],4)
        hex_rep += self.NanonisTCP.to_hex(steps,2)
        hex_rep += self.NanonisTCP.to_hex(group,4)
        hex_rep += self.NanonisTCP.to_hex(wait_until_finished,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def StartClosedLoop(self,abs_rel,target_x,target_y,target_z,wait_until_finished):
        """
        Moves the coarse positioning device (motor, piezo actuator...) in 
        closed loop. This is not supported by all motor control modules

        Parameters
        ----------
        abs_rel             : selects if moving in relative or absolute mode
                              False: relative
                              True:  absolute
        target_x            : target position to move in x (m)
        target_y            : target position to move in y (m)
        target_z            : target position to move in z (m)
        wait_until_finished : defines if this function returns immediately or
                              after reaching destination
                              False: Return immediately
                              True:  Wait until position is reached before 
                                     returning

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.StartClosedLoop', body_size=32)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(abs_rel,4)
        hex_rep += self.NanonisTCP.float64_to_hex(target_x)
        hex_rep += self.NanonisTCP.float64_to_hex(target_y)
        hex_rep += self.NanonisTCP.float64_to_hex(target_z)
        hex_rep += self.NanonisTCP.to_hex(wait_until_finished,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def StopMove(self):
        """
        Stops the motor motion

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.StopMove', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)
    
    def PosGet(self):
        """
        Returns the positions of the motor control module.

        Returns
        -------
        x_pos : the x position (m)
        y_pos : the y position (m)
        z_pos : the z position (m)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.PosGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(24)
        
        x_pos = self.NanonisTCP.hex_to_float64(response[0:8])
        y_pos = self.NanonisTCP.hex_to_float64(response[8:16])
        z_pos = self.NanonisTCP.hex_to_float64(response[16:24])
        
        return [x_pos,y_pos,z_pos]
    
    def StepCounterGet(self,reset_x=False,reset_y=False,reset_z=False):
        """
        Retuns the step counter values of X, Y, Z.
        
        This function also allows to reset the step counters after reading 
        their values through inputs reset_x,reset_y,reset_z

        Parameters
        ----------
        reset_x : reset the x counter (True: reset, False: don't reset)
        reset_y : reset the y counter (True: reset, False: don't reset)
        reset_z : reset the z counter (True: reset, False: don't reset)

        Returns
        -------
        counter_x : x step counter value
        counter_y : y step counter value
        counter_z : z step counter value

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.StepCounterGet', body_size=12)
        
        ## Arguments
        hex_rep += self.NanonisTCP.to_hex(reset_x,4)
        hex_rep += self.NanonisTCP.to_hex(reset_y,4)
        hex_rep += self.NanonisTCP.to_hex(reset_z,4)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(12)
        
        counter_x = self.NanonisTCP.hex_to_int32(response[0:4])
        counter_y = self.NanonisTCP.hex_to_int32(response[4:8])
        counter_z = self.NanonisTCP.hex_to_int32(response[8:12])
        
        return [counter_x,counter_y,counter_z]
    
    def FreqAmpGet(self):
        """
        Returns the frequency (Hz) and amplitude (V) of the motor control 
        module.
        
        This function is only available for PD5, PMD4, and Attocube ANC150
        devices.


        Returns
        -------
        frequency : frequency (Hz)
        amplitude : amplitude (V)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.FreqAmpGet', body_size=0)
        
        self.NanonisTCP.send_command(hex_rep)
        
        response = self.NanonisTCP.receive_response(8)
        
        frequency = self.NanonisTCP.hex_to_float32(response[0:4])
        amplitude = self.NanonisTCP.hex_to_float32(response[4:8])
        
        return [frequency,amplitude]
        
    def FreqAmpSet(self,frequency,amplitude,axis=0):
        """
        Sets the frequency (Hz) and amplitude (V) of the motor control

        Parameters
        ----------
        frequency : frequency (Hz)
        amplitude : amplitude (V)

        """
        ## Make Header
        hex_rep = self.NanonisTCP.make_header('Motor.FreqAmpSet', body_size=10)
        
        ## Arguments
        hex_rep += self.NanonisTCP.float32_to_hex(frequency)
        hex_rep += self.NanonisTCP.float32_to_hex(amplitude)
        hex_rep += self.NanonisTCP.to_hex(axis,2)
        
        self.NanonisTCP.send_command(hex_rep)
        
        self.NanonisTCP.receive_response(0)


# -*- coding: utf-8 -*-
"""
Created on Tue Apr 26 09:33:32 2022

@author: jced0001
"""
import numpy as np
class Scan:
    """
    Nanonis Scan Module
    """
    def __init__(self, nanonisTCP):
        self.nanonisTCP = nanonisTCP
    
    def Action(self, scan_action, scan_direction="up"):
        """
        Sarts, stops, pauses, or resumes a scan

        Parameters
        scan_action     : one of "start", "stop", "pause", "resume"
        scan_direction  : "up" or "down"
        
        """
        action_dict = {"start" : 0,
                       "stop"  : 1,
                       "pause" : 2,
                       "resume": 3,
                       "down"  : 0,
                       "up"    : 1
                       }
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.Action', body_size=6)
        
        ## arguments
        hex_rep += self.nanonisTCP.to_hex(action_dict[scan_action],2)
        hex_rep += self.nanonisTCP.to_hex(action_dict[scan_direction],4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
    
    def WaitEndOfScan(self,timeout=-1):
        """
        Waits for the end-of-scan
        This function returns only when an end-of-scan or timeout occurs (which
        ever occurs first)

        Parameters
        timeout : timeout in ms. if -1, it waits indefinitely

        Returns
        timeout_status : 1: function timed out. 0: didn't time out
        

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.WaitEndOfScan', body_size=4)
        
        ## arguments
        hex_rep += self.nanonisTCP.to_hex(timeout,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response()
        
        timeout_status = self.nanonisTCP.hex_to_uint32(response[0:4]) > 0
        file_path_size = self.nanonisTCP.hex_to_uint32(response[4:8])
        file_path      = response[8:8+file_path_size].decode()
        
        self.nanonisTCP.check_error(response, 8+file_path_size)
        
        return [timeout_status, file_path_size, file_path]
    
    def FrameSet(self,x,y,w,h,angle=0):
        """
        Configures the scan frame position and dimensions
        
        Parameters
        x     : centre x coordinate
        y     : centre y coordinate
        w     : scan frame width
        h     : scan frame height
        angle : scan frame angle (degrees). angle > 0: clockwise
        
        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.FrameSet', body_size=20)
        
        ## arguments
        hex_rep += self.nanonisTCP.float32_to_hex(x)
        hex_rep += self.nanonisTCP.float32_to_hex(y)
        hex_rep += self.nanonisTCP.float32_to_hex(w)
        hex_rep += self.nanonisTCP.float32_to_hex(h)
        hex_rep += self.nanonisTCP.float32_to_hex(angle)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
        
    def FrameGet(self):
        """
        Returns the scan frame position and dimensions

        Returns
        x     : centre x coordinate
        y     : centre y coordinate
        w     : scan frame width
        h     : scan frame height
        angle : scan frame angle (degrees). angle > 0: clockwise 

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.FrameGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response(20)
        
        x     = self.nanonisTCP.hex_to_float32(response[0:4])
        y     = self.nanonisTCP.hex_to_float32(response[4:8])
        w     = self.nanonisTCP.hex_to_float32(response[8:12])
        h     = self.nanonisTCP.hex_to_float32(response[12:16])
        angle = self.nanonisTCP.hex_to_float32(response[16:20])
        
        return [x,y,w,h,angle]
    
    def BufferSet(self,channel_indexes=None,pixels=None,lines=None):
        """
        Configures the scan buffer parameters

        Parameters
        num_channels    : number of recorded channels.
        channel_indexes : indexes of recorded channels (see signals manager or
                          use Signals.InSlotsGet function)
        pixels          : number of pixels per line. forced to a multiple of 16
        lines           : number of scan lines

        """
        _, buf_channel_indexes, buf_pixels, buf_lines = self.BufferGet()        # Get the values currently in the buffer
        
        if(not channel_indexes): channel_indexes = buf_channel_indexes          # Keep the original value if we don't want to change it in this call
        if(not pixels):          pixels = buf_pixels                            # Keep the original value if we don't want to change it in this call
        if(not lines):           lines = buf_lines                              # Keep the original value if we don't want to change it in this call
            
        num_channels = len(channel_indexes)                                     # num_channels is needed as an argument
        body_size    = 12 + 4*num_channels                                      # Variable body size depending on number of channels. each channel index is a float32 thus 4 bytes
        
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.BufferSet', body_size=body_size)
        
        ## arguments
        hex_rep += self.nanonisTCP.to_hex(num_channels,4)
        for c in channel_indexes:
            hex_rep += self.nanonisTCP.to_hex(c,4)
        hex_rep += self.nanonisTCP.to_hex(pixels,4)
        hex_rep += self.nanonisTCP.to_hex(lines,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
        
    def BufferGet(self):
        """
        Returns the scan buffer parameters

        Returns
        num_channels    : number of recorded channels.
        channel_indexes : indexes of recorded channels (see signals manager or
                          use Signals.InSlotsGet function)
        pixels          : number of pixels per line. forced to a multiple of 16
        lines           : number of scan lines

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.BufferGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response()
        
        idx = 0
        channel_indexes = []
        num_channels = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        for c in range(num_channels):
            idx += 4
            channel_indexes.append(self.nanonisTCP.hex_to_int32(response[idx:idx+4]))
        
        idx += 4
        pixels = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        
        idx += 4
        lines = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        
        # Receive Response (check errors)
        idx += 4
        self.nanonisTCP.check_error(response, idx)                              # Makes no sense to check if there's error at this point, but how else do we check?
        
        return [num_channels,channel_indexes,pixels,lines]
        
    def PropsSet(self,continuous_scan=0,bouncy_scan=0,autosave=0,series_name="%y%m%d_%H-%M-%S_SPM",comment=""):
        """
        Configures some of the scan parameters

        Parameters
        continuous_scan : sets whether the scan continues or stops when a frame
                          has been completed.
                          0: no change (leave previous setting)
                          1: turn on
                          2: turn off
                          
        bouncy_scan     : sets whether the scan direction changes when a frame
                          has been completed.
                          0: no change (leave previous setting)
                          1: turn on (scan direction changes each EOS)
                          2: turn off (scan direction doesn't change each EOS)
            
        autosave        : defines the save behaviour when a frame has been
                          completed.
                          0: no change (leave previous setting)
                          1: save all
                          2: save next only
                          3: turn off (save none)
        series_name     : is the base name used for the saved images
        
        comment         : is the comment saved in the file

        """
        series_name_size = int(len(self.nanonisTCP.string_to_hex(series_name))/2)
        comment_size     = int(len(self.nanonisTCP.string_to_hex(comment))/2)
        body_size = 20 + series_name_size + comment_size
        
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.PropsSet', body_size=body_size)
        
        ## arguments
        hex_rep += self.nanonisTCP.to_hex(continuous_scan,4)
        hex_rep += self.nanonisTCP.to_hex(bouncy_scan,4)
        hex_rep += self.nanonisTCP.to_hex(autosave,4)
        hex_rep += self.nanonisTCP.to_hex(series_name_size,4)
        if(series_name_size > 0):
            hex_rep += self.nanonisTCP.string_to_hex(series_name)
        hex_rep += self.nanonisTCP.to_hex(comment_size,4)
        if(comment_size > 0):
            hex_rep += self.nanonisTCP.string_to_hex(comment)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
        
    def PropsGet(self):
        """
        Returns some of the scan parameters
        
        Returns
        continuous_scan : sets whether the scan continues or stops when a frame
                          has been completed.
                          0: off
                          1: on
                          
        bouncy_scan     : sets whether the scan direction changes when a frame
                          has been completed.
                          0: off (scan direction doesn't change each EOS)
                          1: on (scan direction changes each EOS)
            
        autosave        : defines the save behaviour when a frame has been
                          completed.
                          0: save all
                          1: save next only
                          2: off (save none)
        series_name     : is the base name used for the saved images
        
        comment         : is the comment saved in the file

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.PropsGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response()
        
        continuous_scan = self.nanonisTCP.hex_to_uint32(response[0:4])
        bouncy_scan     = self.nanonisTCP.hex_to_uint32(response[4:8])
        autosave        = self.nanonisTCP.hex_to_uint32(response[8:12])
        
        series_name = ""
        series_name_size = self.nanonisTCP.hex_to_int32(response[12:16])
        if(series_name_size > 0):
            series_name = response[16:16+series_name_size].decode()
        
        comment = ""
        idx = 16+series_name_size
        comment_size = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        if(comment_size > 0):
            idx += 4
            comment = response[idx:idx+comment_size].decode()
        
        return [continuous_scan,bouncy_scan,autosave,series_name,comment]
        
    def SpeedSet(self,fwd_speed=0,bwd_speed=0,fwd_line_time=0,bwd_line_time=0,const_param=-1,speed_ratio=0):
        """
        Configures the scan speed parameters. Best practice here is to define
        at least (fwd_speed or fwd_line_time)
               + (corresponding bwd param or speed_ratio)
               
        e.g like one of these:
            
            SpeedSet(fwd_speed=150e-9,bwd_speed=400e-9)
            SpeedSet(fwd_line_time=1,bwd_line_time=0.25)
            SpeedSet(fwd_speed=150e-9,speed_ratio=2)
            SpeedSet(fwd_line_time=1,speed_ratio=2)
        
        Parameters
        fwd_speed       : forward linear speed (m/s)
        bwd_speed       : backward linear speed (m/s)
        fwd_line_time   : forward time per line
        bwd_line_time   : backward time per line
        const_param     : defines which speed parameter to keep constant
                          0: No change (leave setting as is)
                          1: keeps linear speed constant
                          2: keeps time per line constant
        speed_ratio     : defines the backward tip speed relative to the
                          forward tip speed

        """
        buf_fwd_speed, buf_bwd_speed, buf_fwd_line_time, buf_bwd_line_time, buf_const_param, buf_speed_ratio = self.SpeedGet()
        
        if(const_param < 0):                                                    # If const_param is not provided, figure out what it should be based on inputs
            if(fwd_speed and not fwd_line_time):  const_param = 1               # If speed is provided and time is not, then assume we want to set speed
            if(not fwd_speed and fwd_line_time):  const_param = 2               # If time is provided and speed is not, then assume we want to set time
        
        if(const_param < 0):
            if(bwd_speed and not bwd_line_time):  const_param = 1               # If speed is provided and time is not, then assume we want to set speed
            if(not bwd_speed and bwd_line_time):  const_param = 2               # If time is provided and speed is not, then assume we want to set time
                        
        if(bwd_speed and fwd_speed and not speed_ratio):                        # If bwd_speed and fwd_speed have been provided, then set the speed_ratio.)
            speed_ratio = bwd_speed/fwd_speed                                   # Nanonis Speed module seems to do literally nothing with the bwd parameters... it just used the speed_ratio to calc bwd from fwd
        
        if(bwd_line_time and fwd_line_time and not speed_ratio):                # If bwd_line_time and fwd_line_time have been provided...
            speed_ratio = fwd_line_time/bwd_line_time                           # then set the speed ratio similar to above. If speed and time have been provided then the ratio is calculated based on time
        
        # Could add more functionality here to accommodate for other combinations of inputs. Not going to do that unless necessary
        
        if(not fwd_speed):      fwd_speed     = buf_fwd_speed
        if(not bwd_speed):      bwd_speed     = buf_bwd_speed
        if(not fwd_line_time):  fwd_line_time = buf_fwd_line_time
        if(not bwd_line_time):  bwd_line_time = buf_bwd_line_time
        if(not speed_ratio):    speed_ratio   = buf_speed_ratio
        
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.SpeedSet', body_size=22)
        
        ## arguments
        hex_rep += self.nanonisTCP.float32_to_hex(fwd_speed)
        hex_rep += self.nanonisTCP.float32_to_hex(bwd_speed)
        hex_rep += self.nanonisTCP.float32_to_hex(fwd_line_time)
        hex_rep += self.nanonisTCP.float32_to_hex(bwd_line_time)
        hex_rep += self.nanonisTCP.to_hex(const_param,2)
        hex_rep += self.nanonisTCP.float32_to_hex(speed_ratio)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response (check errors)
        self.nanonisTCP.receive_response(0)
        
    def SpeedGet(self):
        """
        Returns the scan speed parameters

        Returns
        fwd_speed       : forward linear speed (m/s)
        bwd_speed       : backward linear speed (m/s)
        fwd_line_time   : forward time per line
        bwd_line_time   : backward time per line
        const_param     : defines which speed parameter to keep constant
                          0: No change (leave setting as is)
                          1: keeps linear speed constant
                          2: keeps time per line constant
        speed_ratio     : defines the backward tip speed relative to the
                          forward tip speed

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.SpeedGet', body_size=0)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response()
        
        fwd_speed     = self.nanonisTCP.hex_to_float32(response[0:4])
        bwd_speed     = self.nanonisTCP.hex_to_float32(response[4:8])
        fwd_line_time = self.nanonisTCP.hex_to_float32(response[8:12])
        bwd_line_time = self.nanonisTCP.hex_to_float32(response[12:16])
        const_param   = self.nanonisTCP.hex_to_uint16(response[16:18])
        speed_ratio   = self.nanonisTCP.hex_to_float32(response[18:22])
        
        return [fwd_speed,bwd_speed,fwd_line_time,bwd_line_time,const_param,speed_ratio]
    
    def FrameDataGrab(self,channel_index,data_direction):
        """
        Returns the scan data of the selected frame

        Parameters
        channel_index   : selects which channel to get the data from. The 
                          channel must be one of the acquired channels. The 
                          list of acquired channels while scanning can be 
                          configured by the function Scan.BufferSet or read by
                          the function Scan.BufferGet
        data_direction : selects the data direction to be read.
                         0: backward
                         1: forward

        Returns
        -------
        None.

        """
        ## Make Header
        hex_rep = self.nanonisTCP.make_header('Scan.FrameDataGrab', body_size=8)
        
        ## arguments
        hex_rep += self.nanonisTCP.to_hex(channel_index,4)
        hex_rep += self.nanonisTCP.to_hex(data_direction,4)
        
        self.nanonisTCP.send_command(hex_rep)
        
        # Receive Response
        response = self.nanonisTCP.receive_response()
        
        idx = 0
        channel_name_size = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        
        idx += 4
        channel_name = response[idx:idx+channel_name_size].decode()
        
        idx += channel_name_size
        scan_data_rows = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        
        idx += 4
        scan_data_columns = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        
        scan_data = np.empty((scan_data_rows,scan_data_columns))
        for i in range(scan_data_rows):
            for j in range(scan_data_columns):
                idx += 4
                scan_data[i,j] = self.nanonisTCP.hex_to_float32(response[idx:idx+4])
            
        idx += 4
        scan_direction = self.nanonisTCP.hex_to_int32(response[idx:idx+4])
        scan_direction = ["down","up"][scan_direction]
        
        return [channel_name,scan_data,scan_direction]